% Copyright (C) <2012>  <Aydin Varol, Mathieu Salzmann, Pascal Fua and Raquel Urtasun>
% 
% Implementation of CLVM method described in:
% A. Varol, M. Salzmann, P. Fua and R. Urtasun, A Constrained Latent Variable Model, 
% Conference on Computer Vision and Pattern Recognition, June 2012.
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Aydin Varol, CVLab-EPFL, 2012.
% aydin.varol@gmail.com, www.aydinvarol.com


clc
clear all
close all



%% set seed
% initialization
rand('seed',123456);
randn('seed',123456);

%% config
   
% dataset
gamma= 1e-3;
alpha = 1e-3;

dataset = 'mesh2x2';
 
resPath = 'res.mat';
saveResultsP = 0;

% number of examples
numTests = 10;
numData =  10;
numValid = 10;
numTrial = 5;

% kern type
kernType = 'poly'; % poly or rbf
normalizeX = 1;

% strategy
trainSum = 0; % train with all the constraints

% type of constraints
squaredConstP = 1;

% optimization settings
vert2vertErrorP = 1;
displayRunDetailsP = 1;
pickBestIterP = 1;
numIt = 10;

if(saveResultsP)
   resFile = resPath;  
end


%% load data (LVM)

[Xall Yall Y Y0 Mesh My Sy dataIndsAll validIndsAll testInds meanX varX] = loadDataPredictions(dataset,numData,numValid,numTests,numTrial,normalizeX);


%% make constraints

if(strcmp(dataset(1:4),'mesh'))    
    if(trainSum==0)
        constrFunHandle = @(Y)LengthConstraintsMex(Y,(Mesh.Edges)',Mesh.Lengths,squaredConstP);
        constrFunHandleActualLength = @(Y)LengthConstraintsMex(Y,(Mesh.Edges)',Mesh.Lengths,0);
        [FQ c] = makeEdgeLengthConstMatrix(Mesh.Edges',Mesh.Lengths,Mesh.NVerts);                
    else        
        error('not implemented yet');                
    end    
end


%% global output definitions

REGP = zeros(length(numData),length(numValid));
REC = zeros(length(numData),length(numValid));

CEGP = zeros(length(numData),length(numValid));
CEC = zeros(length(numData),length(numValid));

STDREGP = zeros(length(numData),length(numValid));
STDREC = zeros(length(numData),length(numValid));

STDCEGP = zeros(length(numData),length(numValid));
STDCEC = zeros(length(numData),length(numValid));


%% training constraint and unconstraint (GP) cases

for nd = 1 : length(numData)
    
    for nv = 1 : length(numValid)
        
        % local output definitions
        if(1)
            localREGP = zeros(numTrial,1);            
            localREC = zeros(numTrial,1);
            
            localCEGP = zeros(numTrial,1);
            localCEC = zeros(numTrial,1);            
        end                        
        
                
        % trials
        for ti = 1 : numTrial
            
            fprintf('\nN=%d, V=%d, trial=%d\n',numData(nd),numValid(nv),ti);
            
            dataInds = dataIndsAll{nd}(ti,:);
            
            Xlabeled = Xall(dataInds,:);
            Ylabeled = Y(dataInds,:);
            
            validInds = validIndsAll{nv}(ti,:);
            Xunlabeled = Xall(validInds,:);
            
            Xtest = Xall(testInds,:);
            Ytest = Yall(testInds,:);
            
            % set-up the kernel
            if(strcmp(kernType,'rbf'))
                GPKern = {'rbf','white'};
                opts = cell(2,1);
                invWidth = 1/mean(mean(dist2(Xlabeled,Xlabeled)));
                %invWidth = 1/mean(mean(dist2([Xlabeled;Xunlabeled],[Xlabeled;Xunlabeled])));
                opts{1} = 1/mean(mean(dist2(Xlabeled,Xlabeled)));
                opts{2} = gamma;
                myKern = @(X1,X2) computeRbfKern(X1,X2,invWidth);
            elseif(strcmp(kernType,'poly'))
                Xlabeled = applyQuadPhi(Xlabeled);
                Xunlabeled = applyQuadPhi(Xunlabeled);
                Xtest = applyQuadPhi(Xtest);
                GPKern = {'lin','white'};
                myKern = @computeLinKern;
                opts = cell(2,1);
                opts{1} = 1;
                opts{2} = gamma;
            else
                error('unknown kernel type');
            end
            
            
            %%%%%%%%%%%%%%%%%%% TRAININGS & PREDICTIONS %%%%%%%%%%%%%%%%%%
                        
            % 1. training unconstrained GP (primal)
            WUC = learnUnconstraintW(Xlabeled,Ylabeled,gamma);
            YGP = predictW(WUC,Xtest,My,Sy);
            
            % 2. training constrained model (primal)            
            WC = learnConstraintW(Xlabeled,Ylabeled,Xunlabeled,My,Sy,gamma,FQ,c,WUC,numIt,pickBestIterP,displayRunDetailsP);
            YC = predictW(WC,Xtest,My,Sy);    
            
                                    
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%                        
            
            %%%%%%%%%%%%%%%%% ERRORS %%%%%%%%%%%%%%%%%%%%%%
            % 1. errors unconstrained GP
            localREGP(ti) =  mean(computeReconstructionError(YGP,Ytest,vert2vertErrorP));
            localCEGP(ti) = mean(computeConstError(YGP,constrFunHandleActualLength));
            
                        
            % 2. errors on constrained model            
            localREC(ti) =  mean(computeReconstructionError(YC,Ytest,vert2vertErrorP));
            localCEC(ti) = mean(computeConstError(YC,constrFunHandleActualLength));                        
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                   
            
        end %end trials
                
        % average errors over trials
        REGP(nd,nv) = mean(localREGP);
        REC(nd,nv) = mean(localREC);
        
        CEGP(nd,nv) = mean(localCEGP);        
        CEC(nd,nv) = mean(localCEC);
        
        STDREGP(nd,nv) = std(localREGP);
        STDREC(nd,nv) = std(localREC);
        
        STDCEGP(nd,nv) = std(localCEGP);      
        STDCEC(nd,nv) = std(localCEC);        
        
    end
end


%% displays

REGP
REC

CEGP
CEC



%% global saves
if(saveResultsP)
    save(resFile,'allModelsGP','Xall','Yall','Y','Y0','Mesh','My','Sy','dataIndsAll','validIndsAll','testInds','meanX','varX',...
        'alpha','gamma','dataset','resPath','kernType',...
        'trainSum','squaredConstP','vert2vertErrorP','pickBestIterP','numIt',...       
        '-regexp', '^RE|^CE|^STDRE|^STDCE',...
        '-regexp', '^allModels');
end

